/* EmblSeqImpl.java
 *
 * created: Mon May 10 1999
 *
 * This file is part of Artemis
 *
 * Copyright (C) 1999  Genome Research Limited
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/flatimpl/EmblSeqImpl.java,v 1.1.1.1 1999/05/19 10:41:48 kmr Exp $
 */

import uk.ac.sanger.pathogens.*;
import uk.ac.sanger.pathogens.embl.*;

import nsdb.*;
import type.*;
import seqdb.*;

import java.io.*;

/**
 *  EmblSeqImpl class
 *
 *  @author Kim Rutherford
 *  @version $Id: EmblSeqImpl.java,v 1.1.1.1 1999/05/19 10:41:48 kmr Exp $
 **/

public class EmblSeqImpl extends _EmblSeqImplBase {
  /**
   *
   **/
  public EmblSeqImpl (final File file) {
    try {
      entry = new DocumentEntry (new FileDocument (file));
    } catch (IOException e) {
      System.err.println (e);
    } catch (InvalidRelationException e) {
      System.err.println (e);
    }
  }

  //
  // IDL:nsdb/NucSeq/getSeq:1.0
  //
  /**
   * retrieve string representation of nucleotide sequence. Each character
   * in the string is a
   *   <href="http://www.ebi.ac.uk/ebi_docs/embl_db/ft/iupac_codes.html">
   *   IUPAC nucleotide base code</A>
   * <P>
   * This method can be used instead of the getAnySeq ()
   * method in seqdb::BioSeq
   * @see seqdb::BioSeq
   **/
  public String getSeq () {
    return entry.getSequence ().getSubSequence (1, entry.getSequence ().length ());
  }

  //
  // IDL:nsdb/NucSeq/getCheckSum:1.0
  //
  /**
   * Checksum on sequence to allow validation.
   *
   **/
  public int getCheckSum () {
    return 0;
  }

  //
  // IDL:nsdb/NucSeq/getTopology:1.0
  //
  /**
   * topology of the nucleotide sequence
   * @see meta::nsdb
   **/
  public String getTopology () {
    return "";
//    throw new Error ("internal error");
  }

  //
  // IDL:nsdb/NucSeq/getMoleculeType:1.0
  //
  /**
   * molecule type of the nucleotide sequence
   * @see meta::nsdb
   **/
  public String getMoleculeType () {
    throw new Error ("internal error");
  }

  //
  // IDL:nsdb/NucSeq/getNucFeatures:1.0
  //
  /**
   * retrieve sequence of NucFeatureList associated with
   * the nucleotide sequence.
   * A sequence has <I>ownership</I> of all these
   * features. It is possible on the other hand that
   * features, owned by another sequence, reference
   * the current sequence. Currently there is no way to find out.
   * @raises type::NoResult if no features are owned by the sequence
   **/
  public NucFeature [] getNucFeatures ()
      throws NoResult {
    final FeatureVector features = entry.getAllFeatures ();

    final NucFeature [] nuc_features = new NucFeature [features.size ()];

    for (int i = 0 ; i < features.size () ; ++i) {
      nuc_features[i] = new NucFeatureImpl (features.elementAt (i));
    }
    
    return nuc_features;
  }

  //
  // IDL:nsdb/NucSeq/getLocalLocation:1.0
  //
  /**
   * A location of a NucFeature can span multiple sequences.
   * If only the location of a feature relevant to the current
   * sequence is required, this method will calculate that.
   * @parm nuc_feature Feature from which location needs to be
   *   intersected with the current sequence.
   * @raises type::InvalidRelation if the nuc_feature
   *   is not associated to the current sequence
   **/
  public nsdb.Location getLocalLocation (final NucFeature nuc_feature)
      throws InvalidRelation {
    throw new Error ("internal error");
  }

  //
  // IDL:nsdb/NucSeq/getOrganisms:1.0
  //
  /**
   * organism(s) from which the NucSeq was obtained.
   * If the sequence is chimeric, multiple organisms will be returned.
   * For each organism, there should be a source feature associated
   * with the current sequence. This source feature has a location
   * defining which part of the sequence was derived from the specified
   * organism.
   * This method provides a shortcut to:
   * invoke getNucFeaturesByKey("source"), invoke getQualifiers () on each
   * source feature and loop through qualifiers to find DbXref.
   * @raises type::NoResult if no source features are associated with
   * the sequence (should never happen).
   **/
  public DbXref [] getOrganisms ()
      throws NoResult {
    return new DbXref [0];
//    throw new Error ("internal error");
  }

  //
  // IDL:nsdb/NucSeq/getNucFeaturesByKey:1.0
  //
  /**
   * retreive all features of a specific kind (FeatureKey)
   * @parm key Type of features to be retrieved
   * @raises type::NoResult if the sequence has no associated features
   *     of the requested
   * @raises type::InvalidArgumentValue if the key is not a
   * <A href="http://www.ebi.ac.uk/ebi_docs/embl_db/ft/feature_table.html">
   * valid feature key</A>
   * @see NucFeature
   **/
  public NucFeature [] getNucFeaturesByKey (final String key)
      throws NoResult, InvalidArgumentValue
  {
    throw new Error ("internal error");
  }

  //
  // IDL:nsdb/NucSeq/getSubSeq:1.0
  //
  /**
   * Create nucleotide sequence of the segment specified
   * The first base in the sequence is numbered 1
   *
   * @parm start first base of sub-sequence (inclusive)
   * @parm end last base of sub-sequence (inclusive)
   * @raises type::IndexOutOfRange if start &lt 1
   *   or end &gt length
   **/
  public String getSubSeq (final int start, final int end)
      throws IndexOutOfRange {
    return entry.getSequence ().getSubSequence (start, end);
  }

  //
  // IDL:nsdb/NucSeq/getSubSeqByFeature:1.0
  //
  /**
   * Create nucleotide sequence of the location derived from the specified
   * feature and contained in the current NucSeq.
   * @raises type::InvalidRelation if the nuc_feature
   *   is not associated to the current sequence
   * @raises InexactLocation if an exact sequence cannot be
   *   determined due to the feature's location being inexact
   **/
  public String getSubSeqByFeature (final NucFeature feature)
      throws InvalidRelation, InexactLocation {
    throw new Error ("internal error");
  }

  //
  // IDL:nsdb/NucSeq/getReferenceLocation:1.0
  //
  /**
   * @raises type::InvalidRelation if reference does not exist
   * @raises type::InvalidRelation if reference is not
   *   associated to the current sequence
   * @raises type::NoResult if there is no location associated with this reference
   **/
  public nsdb.Location getReferenceLocation (final String reference_id)
      throws InvalidRelation, NoResult {
    throw new Error ("internal error");
  }

  //
  // IDL:seqdb/SeqInfo/getDescription:1.0
  //
  /**
   * short (one line) description
   **/
  public String getDescription ()
      throws NoResult {
    throw new Error ("internal error");
  }

  //
  // IDL:seqdb/SeqInfo/getKeywords:1.0
  //
  /**
   * sequence of keywords, describing the characteristics of the sequence
   **/
  public String [] getKeywords ()
      throws NoResult {
    throw new Error ("internal error");
  }

  //
  // IDL:seqdb/SeqInfo/getComments:1.0
  //
  /**
   * sequence of comments, describing the characteristics of the sequence
   **/
  public String [] getComments ()
      throws NoResult {
    throw new Error ("internal error");
  }

  //
  // IDL:seqdb/SeqInfo/getDbXrefs:1.0
  //
  /**
   * cross references to other databases containing related or additional
   * information
   **/
  public DbXref [] getDbXrefs ()
      throws NoResult {
    throw new Error ("internal error");
  }

  //
  // IDL:seqdb/SeqInfo/getReferences:1.0
  //
  /**
   * cross references to the EMBL publication database
   * information
   **/
  public DbXref [] getReferences ()
      throws NoResult {
    throw new Error ("internal error");
  }

  //
  // IDL:nsdb/EntryInfo/getEntryName:1.0
  //
  /**
   * Retrieve entry indentifier. More information on the
   * <A href="http://www.ebi.ac.uk/ebi_docs/embl_db/usrman/id_line.html">format</A> of
   * an entry name is available in the
   *  <A href="http://www.ebi.ac.uk/ebi_docs/embl_db/usrman/usrman.html">
   * EMBL User Manual</A>.
   **/
  public String getEntryName () {
    return entry.getName ();
  }

  //
  // IDL:nsdb/EntryInfo/getEntryVersion:1.0
  //
  /**
   * The entry version defines the current version of an EMBL Sequence Database
   * Entry. I.e. the complete set information related to a particular sequence.
   * The Entry version is incremented whenever anything changes in the sequence or it's
   * associated information.
   **/
  public int getEntryVersion () {
    return 0;
  }

  //
  // IDL:nsdb/EntryInfo/getEntryStatus:1.0
  //
  /**
   * Get entry status code.
   * @see meta::nsdb
   **/
  public String getEntryStatus () {
    throw new Error ("internal error");
  }

  //
  // IDL:nsdb/EntryInfo/getRevisions:1.0
  //
  /**
   * Sequence of revisions when Entry was created/changed.
   **/
  public Revision [] getRevisions () {
    throw new Error ("internal error");
  }

  //
  // IDL:nsdb/EntryInfo/getSecondaryIds:1.0
  //
  /**
   * List of secondary accession numbers. I.e. accession numbers
   * of deprecated entries, now merged into the current entry or
   * split over multiple entries, as decribed in the
   * <A href="http://www.ebi.ac.uk/ebi_docs/embl_db/usrman/ac_line.html">
   * AC line</A> documentation in the User manual.
   **/
  public String [] getSecondaryIds () {
    throw new Error ("internal error");
  }

  //
  // IDL:nsdb/EntryInfo/getCountA:1.0
  //
  /**
   * number of adenine
   **/
  public int getCountA () {
    return entry.getSequence ().getACount ();
  }

  //
  // IDL:nsdb/EntryInfo/getCountC:1.0
  //
  /**
   * number of cytosine
   **/
  public int getCountC () {
    return entry.getSequence ().getCCount ();
  }

  //
  // IDL:nsdb/EntryInfo/getCountG:1.0
  //
  /**
   * number of guanine
   **/
  public int getCountG () {
    return entry.getSequence ().getGCount ();
  }

  //
  // IDL:nsdb/EntryInfo/getCountT:1.0
  //
  /**
   * number of thymine (in DNA sequence)
   *    or
   * number of uracil (in RNA sequence)
   **/
  public int getCountT () {
    return entry.getSequence ().getTCount ();
  }

  // IDL:seqdb/BioSeq/getBioSeqId:1.0
  //
  /**
   * retrieve unique identifier
   **/
  public String getBioSeqId() {
    return entry.getName ();
  }

  //
  // IDL:seqdb/BioSeq/getLength:1.0
  //
  /**
   * length (nr of elements) of the biosequence
   **/
  public int getLength() {
    return entry.getSequence ().length ();
  }

  //
  // IDL:seqdb/BioSeq/getAnySeq:1.0
  //
  /**
   * sequence of objects describing the elements in the biosequence
   * This is a generic description. Most subclasses
   * will define more convenient methods for accessing the biosequence.
   * @returns any containing an set of objects. The any should have
   *          a typecode tk_array.
   **/
  public org.omg.CORBA.Any getAnySeq() {
    throw new Error ("internal error");
  }

  //
  // IDL:seqdb/BioSeq/getBioSeqVersion:1.0
  //
  /**
   * Return current version of the BioSeq. returns 0 if versioning is not
   * implemented on the bioseq.
   **/
  public int getBioSeqVersion() {
    return 0;
  }

  private Entry entry;
}

