/* EmblSeqWriterImpl.java
 *
 * created: Fri May 28 1999
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1999  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/flatimpl/EmblSeqWriterImpl.java,v 1.2.2.4 2000/10/20 15:03:28 kmr Exp $
 */

import nsdb.*;
import type.*;
import seqdb.*;

import diana.Options;

import uk.ac.sanger.pathogens.*;
import uk.ac.sanger.pathogens.embl.Location;
import uk.ac.sanger.pathogens.embl.Feature;
import uk.ac.sanger.pathogens.embl.DocumentEntryFactory;
import uk.ac.sanger.pathogens.embl.DocumentEntry;
import uk.ac.sanger.pathogens.embl.InvalidRelationException;
import uk.ac.sanger.pathogens.embl.InvalidKeyException;
import uk.ac.sanger.pathogens.embl.LocationParseException;
import uk.ac.sanger.pathogens.embl.Entry;
import uk.ac.sanger.pathogens.embl.FeatureVector;
import uk.ac.sanger.pathogens.embl.Key;
import uk.ac.sanger.pathogens.embl.Range;
import uk.ac.sanger.pathogens.embl.EntryInformationException;
import uk.ac.sanger.pathogens.embl.EntryInformation;

import java.io.*;
import java.util.*;

/**
 *  The implementation of the EmblSeqWriter interface.
 *
 *  @author Kim Rutherford
 *  @version $Id: EmblSeqWriterImpl.java,v 1.2.2.4 2000/10/20 15:03:28 kmr Exp $
 **/

public class EmblSeqWriterImpl
    extends _EmblSeqWriterImplBase
    implements nsdb.EmblSeqWriter, nsdb.EmblSeq {
  /**
   *  Create a new EmblSeqWriterImpl object from the given File.
   **/
  public EmblSeqWriterImpl (final File db_file)
      throws IOException, InvalidRelationException {
    final Document entry_file = new FileDocument (db_file);
    final EntryInformation entry_information =
      Options.getOptions ().getArtemisEntryInformation ();
    try {
      entry = DocumentEntryFactory.makeDocumentEntry (entry_information,
                                                      entry_file);
    } catch (EntryInformationException e) {
      throw new IOException ("An exception occured while reading: " + e);
    }
  }

  //
  // IDL:nsdb/NucSeq/getSeq:1.0
  //
  /**
   * retrieve string representation of nucleotide sequence. Each character
   * in the string is a
   *   <href="http://www.ebi.ac.uk/ebi_docs/embl_db/ft/iupac_codes.html">
   *   IUPAC nucleotide base code</A>
   * <P>
   * This method can be used instead of the getAnySeq ()
   * method in seqdb::BioSeq
   * @see seqdb::BioSeq
   **/
  public String getSeq () {
    final uk.ac.sanger.pathogens.embl.Sequence sequence = entry.getSequence ();

    if (sequence == null) {
      return "";
    } else {
      return sequence.getSubSequence (1, entry.getSequence ().length ());
    }
  }

  //
  // IDL:nsdb/NucSeq/getCheckSum:1.0
  //
  /**
   * Checksum on sequence to allow validation.
   *
   **/
  public int getCheckSum () {
    return 0;
  }

  //
  // IDL:nsdb/NucSeq/getTopology:1.0
  //
  /**
   * topology of the nucleotide sequence
   * @see meta::nsdb
   **/
  public String getTopology () {
    return "";
//    throw new Error ("internal error");
  }

  //
  // IDL:nsdb/NucSeq/getMoleculeType:1.0
  //
  /**
   * molecule type of the nucleotide sequence
   * @see meta::nsdb
   **/
  public String getMoleculeType () {
    throw new Error ("internal error");
  }

  //
  // IDL:nsdb/NucSeq/getNucFeatures:1.0
  //
  /**
   * retrieve sequence of NucFeatureList associated with
   * the nucleotide sequence.
   * A sequence has <I>ownership</I> of all these
   * features. It is possible on the other hand that
   * features, owned by another sequence, reference
   * the current sequence. Currently there is no way to find out.
   * @throws type::NoResult if no features are owned by the sequence
   **/
  public NucFeature [] getNucFeatures ()
      throws NoResult {
    final FeatureVector features = entry.getAllFeatures ();

    final NucFeature [] nuc_features = new NucFeature [features.size ()];

    for (int i = 0 ; i < features.size () ; ++i) {
      nuc_features[i] =
        NucFeatureWriterImpl.getNucFeatureOfFeature (this,
                                                     features.elementAt (i));

      if (feature_dictionary.get (nuc_features[i]) == null) {
        feature_dictionary.put (nuc_features[i], features.elementAt (i));
      }
    }
    
    return nuc_features;
  }

  //
  // IDL:nsdb/NucSeq/getNucFeatures:1.0
  //
  /**
   * retrieve sequence of NucFeatureList associated with
   * the nucleotide sequence that are within the given range of bases.
   * @throws NoResult if no features are owned by the sequence
   * @throws IndexOutOfRange if either of start_base or end_base is
   *  less than 1 or greater than the length of the sequence.
   */
  public NucFeature [] getNucFeaturesInRange (final int start_base,
                                              final int end_base)
      throws NoResult, IndexOutOfRange {

    if (start_base < 1 || end_base < 1 ||
        entry.getSequence () != null && 
        (start_base > entry.getSequence ().length () ||
         end_base > entry.getSequence ().length ())) {
      throw new IndexOutOfRange ();
    }

    final Range range;

    try {
      if (end_base < start_base) {
        range = new Range (end_base, start_base);
      } else {
        range = new Range (start_base, end_base);
      }
    } catch (OutOfRangeException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    }

    try {
      final FeatureVector features = entry.getFeaturesInRange (range);

      final NucFeature [] nuc_features = new NucFeature [features.size ()];;

      for (int i = 0 ; i < features.size () ; ++i) {
        nuc_features[i] =
          NucFeatureWriterImpl.getNucFeatureOfFeature (this,
                                                       features.elementAt (i));

        if (feature_dictionary.get (nuc_features[i]) == null) {
          feature_dictionary.put (nuc_features[i], features.elementAt (i));
        }
      }

      return nuc_features;
    } catch (OutOfRangeException e) {
      throw new type.IndexOutOfRange ();
    }
  }

  //
  // IDL:nsdb/EmblSeqWriter/getFeatureAtIndex:1.0
  //
  /**
   * Return the ith NucFeature from this Entry.  The feature are returned in
   * a consistent order, sorted by the first base of each Feature.
   * @raises type::IndexOutOfRange if the index is less than 0 or greater
   *   or equal to than the number of features.
   **/
  public NucFeature getFeatureAtIndex(int i)
      throws type.IndexOutOfRange {
    final Feature feature = entry.getFeatureAtIndex (i);

    return NucFeatureWriterImpl.getNucFeatureOfFeature (this, feature);
  }

  //
  // IDL:nsdb/EmblSeqWriter/indexOf:1.0
  //
  /**
   * Return the index of the given Feature.  This does the reverse of
   * getFeatureAtIndex ().  Returns -1 if the given NucFeature is not in
   * this EmblSeq object.
   **/
  public int indexOf (NucFeature nuc_feature) {
    final Feature feature = getFeatureOfNucFeature (nuc_feature);

    if (feature == null) {
      return -1;
    }

    return entry.indexOf (feature);
  }
  
  //
  // IDL:nsdb/NucSeq/getLocalLocation:1.0
  //
  /**
   * A location of a NucFeature can span multiple sequences.
   * If only the location of a feature relevant to the current
   * sequence is required, this method will calculate that.
   * @param nuc_feature Feature from which location needs to be
   *   intersected with the current sequence.
   * @throws type::InvalidRelation if the nuc_feature
   *   is not associated to the current sequence
   **/
  public nsdb.Location getLocalLocation (final NucFeature nuc_feature)
      throws InvalidRelation {
    throw new Error ("internal error");
  }

  //
  // IDL:nsdb/NucSeq/getOrganisms:1.0
  //
  /**
   * organism(s) from which the NucSeq was obtained.
   * If the sequence is chimeric, multiple organisms will be returned.
   * For each organism, there should be a source feature associated
   * with the current sequence. This source feature has a location
   * defining which part of the sequence was derived from the specified
   * organism.
   * This method provides a shortcut to:
   * invoke getNucFeaturesByKey("source"), invoke getQualifiers () on each
   * source feature and loop through qualifiers to find DbXref.
   * @throws type::NoResult if no source features are associated with
   * the sequence (should never happen).
   **/
  public DbXref [] getOrganisms ()
      throws NoResult {
    return new DbXref [0];
//    throw new Error ("internal error");
  }

  //
  // IDL:nsdb/NucSeq/getNucFeaturesByKey:1.0
  //
  /**
   * retreive all features of a specific kind (FeatureKey)
   * @param key Type of features to be retrieved
   * @throws type::NoResult if the sequence has no associated features
   *     of the requested
   * @throws type::InvalidArgumentValue if the key is not a
   * <A href="http://www.ebi.ac.uk/ebi_docs/embl_db/ft/feature_table.html">
   * valid feature key</A>
   * @see NucFeature
   **/
  public NucFeature [] getNucFeaturesByKey (final String key)
      throws NoResult, InvalidArgumentValue
  {
    throw new Error ("internal error");
  }

  //
  // IDL:nsdb/NucSeq/getSubSeq:1.0
  //
  /**
   * Create nucleotide sequence of the segment specified
   * The first base in the sequence is numbered 1
   *
   * @param start first base of sub-sequence (inclusive)
   * @param end last base of sub-sequence (inclusive)
   * @throws type::IndexOutOfRange if start &lt 1
   *   or end &gt length
   **/
  public String getSubSeq (final int start, final int end)
      throws IndexOutOfRange {
    return entry.getSequence ().getSubSequence (start, end);
  }

  //
  // IDL:nsdb/NucSeq/getSubSeqByFeature:1.0
  //
  /**
   * Create nucleotide sequence of the location derived from the specified
   * feature and contained in the current NucSeq.
   * @throws type::InvalidRelation if the nuc_feature
   *   is not associated to the current sequence
   * @throws InexactLocation if an exact sequence cannot be
   *   determined due to the feature's location being inexact
   **/
  public String getSubSeqByFeature (final NucFeature feature)
      throws InvalidRelation, InexactLocation {
    throw new Error ("internal error");
  }

  //
  // IDL:nsdb/NucSeq/getReferenceLocation:1.0
  //
  /**
   * @throws type::InvalidRelation if reference does not exist
   * @throws type::InvalidRelation if reference is not
   *   associated to the current sequence
   * @throws type::NoResult if there is no location associated with this reference
   **/
  public nsdb.Location getReferenceLocation (final String reference_id)
      throws InvalidRelation, NoResult {
    throw new Error ("internal error");
  }

  //
  // IDL:seqdb/SeqInfo/getDescription:1.0
  //
  /**
   * short (one line) description
   **/
  public String getDescription ()
      throws NoResult {
    throw new Error ("internal error");
  }

  //
  // IDL:seqdb/SeqInfo/getKeywords:1.0
  //
  /**
   * sequence of keywords, describing the characteristics of the sequence
   **/
  public String [] getKeywords ()
      throws NoResult {
    throw new Error ("internal error");
  }

  //
  // IDL:seqdb/SeqInfo/getComments:1.0
  //
  /**
   * sequence of comments, describing the characteristics of the sequence
   **/
  public String [] getComments ()
      throws NoResult {
    throw new Error ("internal error");
  }

  //
  // IDL:seqdb/SeqInfo/getDbXrefs:1.0
  //
  /**
   * cross references to other databases containing related or additional
   * information
   **/
  public DbXref [] getDbXrefs ()
      throws NoResult {
    throw new Error ("internal error");
  }

  //
  // IDL:seqdb/SeqInfo/getReferences:1.0
  //
  /**
   * cross references to the EMBL publication database
   * information
   **/
  public DbXref [] getReferences ()
      throws NoResult {
    throw new Error ("internal error");
  }

  //
  // IDL:nsdb/EntryInfo/getEntryName:1.0
  //
  /**
   * Retrieve entry indentifier. More information on the
   * <A href="http://www.ebi.ac.uk/ebi_docs/embl_db/usrman/id_line.html">format</A> of
   * an entry name is available in the
   *  <A href="http://www.ebi.ac.uk/ebi_docs/embl_db/usrman/usrman.html">
   * EMBL User Manual</A>.
   **/
  public String getEntryName () {
    return entry.getName ();
  }

  //
  // IDL:nsdb/EntryInfo/getEntryVersion:1.0
  //
  /**
   * The entry version defines the current version of an EMBL Sequence Database
   * Entry. I.e. the complete set information related to a particular sequence.
   * The Entry version is incremented whenever anything changes in the sequence or it's
   * associated information.
   **/
  public int getEntryVersion () {
    return 0;
  }

  //
  // IDL:nsdb/EntryInfo/getEntryStatus:1.0
  //
  /**
   * Get entry status code.
   * @see meta::nsdb
   **/
  public String getEntryStatus () {
    throw new Error ("internal error");
  }

  //
  // IDL:nsdb/EntryInfo/getRevisions:1.0
  //
  /**
   * Sequence of revisions when Entry was created/changed.
   **/
  public Revision [] getRevisions () {
    throw new Error ("internal error");
  }

  //
  // IDL:nsdb/EntryInfo/getSecondaryIds:1.0
  //
  /**
   * List of secondary accession numbers. I.e. accession numbers
   * of deprecated entries, now merged into the current entry or
   * split over multiple entries, as decribed in the
   * <A href="http://www.ebi.ac.uk/ebi_docs/embl_db/usrman/ac_line.html">
   * AC line</A> documentation in the User manual.
   **/
  public String [] getSecondaryIds () {
    throw new Error ("internal error");
  }

  //
  // IDL:nsdb/EntryInfo/getCountA:1.0
  //
  /**
   * number of adenine
   **/
  public int getCountA () {
    return entry.getSequence ().getACount ();
  }

  //
  // IDL:nsdb/EntryInfo/getCountC:1.0
  //
  /**
   * number of cytosine
   **/
  public int getCountC () {
    return entry.getSequence ().getCCount ();
  }

  //
  // IDL:nsdb/EntryInfo/getCountG:1.0
  //
  /**
   * number of guanine
   **/
  public int getCountG () {
    return entry.getSequence ().getGCount ();
  }

  //
  // IDL:nsdb/EntryInfo/getCountT:1.0
  //
  /**
   * number of thymine (in DNA sequence)
   *    or
   * number of uracil (in RNA sequence)
   **/
  public int getCountT () {
    return entry.getSequence ().getTCount ();
  }

  // IDL:seqdb/BioSeq/getBioSeqId:1.0
  //
  /**
   * retrieve unique identifier
   **/
  public String getBioSeqId() {
    return entry.getName ();
  }

  //
  // IDL:seqdb/BioSeq/getLength:1.0
  //
  /**
   * length (nr of elements) of the biosequence
   **/
  public int getLength() {
    return entry.getSequence ().length ();
  }

  //
  // IDL:seqdb/BioSeq/getAnySeq:1.0
  //
  /**
   * sequence of objects describing the elements in the biosequence
   * This is a generic description. Most subclasses
   * will define more convenient methods for accessing the biosequence.
   * @returns any containing an set of objects. The any should have
   *          a typecode tk_array.
   **/
  public org.omg.CORBA.Any getAnySeq() {
    throw new Error ("internal error");
  }

  //
  // IDL:seqdb/BioSeq/getBioSeqVersion:1.0
  //
  /**
   * Return current version of the BioSeq. returns 0 if versioning is not
   * implemented on the bioseq.
   **/
  public int getBioSeqVersion() {
    return 0;
  }

  //
  // IDL:nsdb/EmblSeqWriter/createNucFeature:1.0
  //
  /**
   * Create a new feature in this EmblSeq object.
   * @param key Type of the feature to be created
   * @param location_string The location of the new NucFeature
   * @throws LocationParse If the location string is not a valid location.
   * @throws type::IndexOutOfRange If any part of the location is beyond the
   *   end of the sequence
   * @throws InvalidKey if the given key is not a possible EMBL key.
   **/
  public NucFeatureWriter createNucFeature(String key_string,
                                           String location_string)
      throws LocationParse, type.IndexOutOfRange, InvalidKey,
             nsdb.ReadOnlyException {
    try {
      final Key key = new Key (key_string);
      final Location location = new Location (location_string);
      final Feature new_feature = entry.createFeature (key, location, null);
      final NucFeatureWriter new_nuc_feature =
        NucFeatureWriterImpl.getNucFeatureOfFeature (this, new_feature);

      feature_dictionary.put (new_nuc_feature, new_feature);

      return new_nuc_feature;
    } catch (uk.ac.sanger.pathogens.ReadOnlyException e) {
      throw new nsdb.ReadOnlyException ();
    } catch (InvalidKeyException e) {
      throw new InvalidKey ();
    } catch (InvalidRelationException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    } catch (LocationParseException e) {
      throw new LocationParse (e.getMessage ());
    } catch (OutOfRangeException e) {
      throw new type.IndexOutOfRange ();
    } catch (EntryInformationException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    }
  }

  //
  // IDL:nsdb/EmblSeqWriter/remove:1.0
  //
  /**
   * Remove the given feature.
   **/
  public void remove(NucFeature nuc_feature)
      throws nsdb.ReadOnlyException {
    final Feature feature = getFeatureOfNucFeature (nuc_feature);

    if (feature == null) {
      System.out.println ("feature: " + feature);
      // removed earlier?
    } else {
      try {
        if (entry.remove (feature)) {
//          feature_dictionary.remove (feature);
        }
      } catch (uk.ac.sanger.pathogens.ReadOnlyException e) {
        throw new nsdb.ReadOnlyException ();
      }
    }
  }

  //
  // IDL:nsdb/EmblSeqWriter/getNucFeatureCount:1.0
  //
  /**
   *  Return the number of features
   **/
  public int getNucFeatureCount () {
    return entry.getFeatureCount ();
  }

  //
  // IDL:nsdb/EmblSeqWriter/commit:1.0
  //
  /**
   *  commit any pending changes to the database immediately.
   **/
  public void commit()
      throws CommitFailed {
    try {
      entry.save ();
    } catch (IOException e) {
      throw new CommitFailed ("IOException: " + e.getMessage ());
    }
  }

  //
  // IDL:nsdb/EmblSeqWriter/getDatestamp:1.0
  //
  /**
   * Return a Datestamp that will be passed to the set methods on the
   * EmblSeqWriter methods and NucFeatureWriter methods.  The object that is
   * returned represents the time when the entry was last changed (the last
   * time a feature was added, removed or changed location).
   **/
  public Datestamp getDatestamp () {
    return datestamp;
  }

  /**
   *  Returns the Feature reference that corresponds to the given
   *  NucFeatureWriter reference.  This method uses feature_dictionary.  This
   *  method will always return the same Feature reference for a given
   *  NucFeatureWriter.
   **/
  private Feature getFeatureOfNucFeature (final NucFeature nuc_feature) {
    final Feature feature =
      (Feature) feature_dictionary.get (nuc_feature);

    return feature;
  }


  /**
   *  Update the datestamp field of this feature so that it contains the
   *  current date/time.
   **/
  void touch () {
    final java.util.Calendar calendar = java.util.Calendar.getInstance ();

    final java.util.Date current_time = calendar.getTime ();

    datestamp = new Datestamp ();

    datestamp.value = (int) (current_time.getTime () / 1000);
  }

  /**
   *  The underlying object of the EmblSeqWriter object.
   **/
  Entry entry;

  /**
   *  The Dictionary is used to find the the embl.Feature reference that
   *  corresponds any given NucFeatureWriter object.  (see
   *  getFeatureOfNucFeature ()).
   **/
  java.util.Hashtable feature_dictionary = new java.util.Hashtable ();

  /**
   *  The date/time this entry was last changed
   **/
  private Datestamp datestamp;
}

