/* NucFeatureWriterImpl.java
 *
 * created: Fri May 28 1999
 *
 * This file is part of Artemis
 *
 * Copyright (C) 1999  Genome Research Limited
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/flatimpl/NucFeatureWriterImpl.java,v 1.3.2.4 2000/09/20 14:26:51 kmr Exp $
 */

import uk.ac.sanger.pathogens.*;
import uk.ac.sanger.pathogens.embl.*;
import uk.ac.sanger.pathogens.embl.Feature;

import type.*;
import seqdb.*;

import nsdb.NucFeaturePackage.QualifierValue_u;

/**
 *  The implementation of the NucFeatureWriter interface.
 *
 *  @author Kim Rutherford
 *  @version $Id: NucFeatureWriterImpl.java,v 1.3.2.4 2000/09/20 14:26:51 kmr Exp $
 **/

public class NucFeatureWriterImpl
    extends nsdb._NucFeatureWriterImplBase {
  /**
   *  Create a new NucFeatureWriterImpl object from an EmblSeqWriterImpl
   *  object and a embl.Feature object.
   **/
  public NucFeatureWriterImpl (final EmblSeqWriterImpl owner,
                               final Feature feature) {
    this.feature = feature;
    this.owner = owner;
    touch ();
  }

  //
  // IDL:nsdb/NucFeatureWriter/set:1.0
  //
  /**
   * Set the key, location and qualifiers of this NucFeature
   * @param key The new feature key
   * @param location The new feature location
   * @exception InvalidKey Thrown if the given key is not a possible EMBL key.
   * @exception type.IndexOutOfRange If any part of the location is out
   *   of range for the sequence.
   * @exception LocationParse Thrown if the location string is not a valid
   *   location.
   * @exception type.InvalidRelation Thrown if one of the qualifiers in this
   *   feature cannot be associated with the given feature key.
   * @exception QualifierParse Thrown if the format of any the qualifiers is
   *   not appropriate for a Qualifier with the given name.  For
   *   example the value part of /codon_start qualifier must be a number: 1,
   *   2 or 3.  Also thrown if a qualifier has value when it should not or
   *   vice versa.
   * @exception InvalidQualifier Thrown if the name of the Qualifier is not a
   *   valid embl qualifier name.
   * @exception nsdb.ReadOnlyException Thrown if this Feature cannot be changed.
   * @exception OutOfDate Thrown if the key has changed since the time given by
   *   datestamp.
   **/
  public void set (nsdb.Datestamp datestamp,
                   String key_string,
                   String location_string,
                   nsdb.NucFeaturePackage.Qualifier[] qualifier_list)
      throws nsdb.InvalidKey, type.IndexOutOfRange,nsdb.LocationParse,
             type.InvalidRelation, nsdb.QualifierParse,
             nsdb.InvalidQualifier, nsdb.OutOfDate,
             nsdb.ReadOnlyException {
    checkStamp (datestamp);


    try {
      final QualifierVector new_qualifiers = new QualifierVector ();

      for (int i = 0 ; i < qualifier_list.length ; ++i) {
        final Qualifier new_qualifier =
          getQualifierFromStruct (qualifier_list [i]);
        new_qualifiers.setQualifier (new_qualifier);
      }

      maybeReAddFeature ();

      final uk.ac.sanger.pathogens.embl.Location new_location =
        new uk.ac.sanger.pathogens.embl.Location (location_string);

      feature.set (new Key (key_string), new_location, new_qualifiers);

      touch ();
      owner.touch ();
    } catch (LocationParseException e) {
      throw new nsdb.LocationParse (e.getMessage ());
    } catch (InvalidKeyException e) {
      throw new nsdb.InvalidKey ();
    } catch (ReadOnlyException e) {
      throw new nsdb.ReadOnlyException ();
    } catch (InvalidRelationException e) {
      throw new InvalidRelation (e.getMessage ());
    } catch (QualifierParseException e) {
      throw new nsdb.QualifierParse ();
    } catch (InvalidQualifierException e) {
      throw new nsdb.InvalidQualifier ();
    } catch (OutOfRangeException e) {
      throw new type.IndexOutOfRange ();
    } catch (EntryInformationException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    }
  }

  //
  // IDL:seqdb/Feature/getKey:1.0
  //
  /**
   * Feature types are defined by a key.
   **/
  public String getKey () {
//    System.out.println ("server getKey ()");

    return feature.getKey ().toString ();
  }

  //
  // IDL:nsdb/NucFeatureWriter/setKey:1.0
  //
  /**
   * Set the key of this NucFeature
   * @param key The new feature key
   * @exception InvalidKey if the given key is not a possible EMBL key.
   * @exception type.InvalidRelation if one of the qualifiers in this
   *   feature cannot be associated with the given feature key.
   * @exception nsdb.ReadOnlyException Thrown if this Feature cannot be changed.
   * @exception OutOfDate If the key has changed since the time given by
   *   datestamp.
   **/
  public void setKey (final nsdb.Datestamp datestamp, final String key)
      throws nsdb.InvalidKey, type.InvalidRelation, nsdb.OutOfDate,
             nsdb.ReadOnlyException {
    checkStamp (datestamp);

    try {
      maybeReAddFeature ();
      feature.setKey (new Key (key));
//      System.out.println ("server setKey (): " + key.toString ());
      touch ();
    } catch (InvalidKeyException e) {
      throw new nsdb.InvalidKey ();
    } catch (ReadOnlyException e) {
      throw new nsdb.ReadOnlyException ();
    } catch (InvalidRelationException e) {
      throw new InvalidRelation (e.getMessage ());
    } catch (EntryInformationException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    }
  }


  //
  // IDL:nsdb/NucFeature/getLocation:1.0
  //
  /**
   * retrieve location of feature.
   * @exception type.NoResult if no location is associated with the feature.
   **/
  public nsdb.FeatureLocation getLocation ()
      throws type.NoResult {
 
    if (feature_location_impl == null) {
      feature_location_impl = new FeatureLocationImpl (getDbFeature ());
    }
  
    return feature_location_impl;
  }

  //
  // IDL:nsdb/NucFeatureWriter/setLocation:1.0
  //
  /**
   * Set the location of this NucFeature
   * @param location The new feature location
   * @exception type.IndexOutOfRange Thrown if any part of the location is out
   *   of range for the sequence.
   * @exception LocationParse If the location string is not a valid location.
   * @exception nsdb.ReadOnlyException Thrown if this Feature cannot be changed.
   * @exception OutOfDate If the location has changed since the time given by
   *   datestamp.
   **/
  public void setLocation (final nsdb.Datestamp datestamp,
                           final String location)
      throws type.IndexOutOfRange, nsdb.LocationParse, nsdb.OutOfDate,
             nsdb.ReadOnlyException {
    checkStamp (datestamp);

    try {
      final uk.ac.sanger.pathogens.embl.Location new_location =
        new uk.ac.sanger.pathogens.embl.Location (location);
//      System.out.println ("server setLocation (): " + location.toString ());
      maybeReAddFeature ();
      feature.setLocation (new_location);
      touch ();
      owner.touch ();
    } catch (LocationParseException e) {
      throw new nsdb.LocationParse (e.getMessage ());
    } catch (ReadOnlyException e) {
      throw new nsdb.ReadOnlyException ();
    } catch (OutOfRangeException e) {
      throw new type.IndexOutOfRange ();
    }
  }

  /**
   * Retrieve sequence of qualifiers.
   * @exception type.NoResult if no qualifiers are associated with
   *    the feature.
   **/
  public nsdb.NucFeaturePackage.Qualifier [] getQualifiers ()
      throws NoResult {
//    System.out.println ("server getQualifiers ()");
    final QualifierVector qualifiers = feature.getQualifiers ();

    final nsdb.NucFeaturePackage.Qualifier [] nuc_qualifiers =
      new nsdb.NucFeaturePackage.Qualifier [qualifiers.size ()];

    for (int i = 0 ; i < qualifiers.size () ; ++i) {
      final Qualifier this_qualifer = qualifiers.elementAt (i);

      nuc_qualifiers [i] = getNucQualifier (this_qualifer);
    }

    return nuc_qualifiers;
  }

  //
  // IDL:nsdb/NucFeature/getQualifier:1.0
  //
  /**
   * retrieve qualifier of a certain type.
   * To find out which qualifier/feature combinations are valid, a client should
   * query the NucFeatureMeta server.
   * @exception type.NoResult if no qualifier of this type is associated with
   *   the feature
   * @exception type.InvalidRelation if the requested qualifier cannot be
   *   associated to the current feature type
   * @see metafeature.NucFeatureMeta
   **/
  public nsdb.NucFeaturePackage.Qualifier getQualifier (String qualifier_name)
      throws NoResult, InvalidRelation {
//    System.out.println ("server getQualifier ()");
    try {
      final Qualifier qualifier =
        getDbFeature ().getQualifierByName (qualifier_name);

      if (qualifier == null) {
        throw new NoResult ();
      } else {
        return getNucQualifier (qualifier);
      }
    } catch (InvalidRelationException e) {
      throw new InvalidRelation ();
    }
  }

  /**
   *  Given an embl.Qualifier return a corresponding
   *  nsdb.NucFeaturePackage.Qualifier object.
   **/
  private nsdb.NucFeaturePackage.Qualifier getNucQualifier (Qualifier arg) {

    final String name = arg.getName ();
    final StringVector values = arg.getValues ();

    final QualifierValue_u [] nuc_qualifier_values =
      new QualifierValue_u [values == null ? 0 : values.size ()];

    for (int value_index = 0 ;
         value_index < nuc_qualifier_values.length ;
         ++value_index) {
      final QualifierValue_u qualifier_value_u = new QualifierValue_u ();

      nuc_qualifier_values [value_index] = qualifier_value_u;
      qualifier_value_u.text (values.elementAt (value_index));
    }

    return new nsdb.NucFeaturePackage.Qualifier (name, nuc_qualifier_values);
  }

  //
  // IDL:nsdb/NucFeatureWriter/setQualifiers:1.0
  //
  /**
   * Set the qualifiers of this feature, replacing the current qualifiers.
   * @exception InvalidRelationException if this Feature cannot
   *   contain one of the given qualifiers.
   * @exception QualifierParse if the format of any the qualifiers is not
   *   appropriate for a Qualifier with the given name.  For
   *   example the value part of /codon_start qualifier must be a number: 1,
   *   2 or 3.  Also thrown if a qualifier has value when it should not or
   *   vice versa.
   * @exception InvalidQualifier if the name of the Qualifier is not a
   *   valid embl qualifier name.
   * @exception nsdb.ReadOnlyException Thrown if this Feature cannot be changed.
   * @exception OutOfDate if any of the qualifiers has changed since the time
   *   given by datestamp.
   **/
  public void setQualifiers (nsdb.Datestamp datestamp,
                             nsdb.NucFeaturePackage.Qualifier[] qualifier_list)
      throws type.InvalidRelation, nsdb.QualifierParse,
             nsdb.InvalidQualifier, nsdb.OutOfDate,
             nsdb.ReadOnlyException {
    checkStamp (datestamp);

    try {
      final QualifierVector new_qualifiers = new QualifierVector ();

      for (int i = 0 ; i < qualifier_list.length ; ++i) {
        final Qualifier new_qualifier =
          getQualifierFromStruct (qualifier_list [i]);
        new_qualifiers.setQualifier (new_qualifier);
      }

      maybeReAddFeature ();
      feature.setQualifiers (new_qualifiers);
//      System.out.println ("server setQualifiers (): ");
      touch ();
    } catch (ReadOnlyException e) {
      throw new nsdb.ReadOnlyException ();
    } catch (InvalidRelationException e) {
      throw new InvalidRelation (e.getMessage ());
    } catch (QualifierParseException e) {
      throw new nsdb.QualifierParse ();
    } catch (InvalidQualifierException e) {
      throw new nsdb.InvalidQualifier ();
    } catch (EntryInformationException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    }
  }

  //
  // IDL:nsdb/NucFeatureWriter/setQualifier:1.0
  //
  /**
   * Add the given Qualifier to this Feature.  If this Feature contains a
   * Qualifier with the same name as the new Qualifier it will be replaced.
   * @param qualifier The new qualifier to add.
   * @exception InvalidRelationException if this Feature cannot
   *   contain the given Qualifier.
   * @exception QualifierParse if the format of the qualifier is not
   *   appropriate for a Qualifier with the given name.  For
   *   example the value part of /codon_start qualifier must be a number: 1,
   *   2 or 3.  Also thrown if a qualifier has value when it should not or
   *   vice versa.
   * @exception InvalidQualifier if the name of the Qualifier is not a
   *   valid embl qualifier name.
   * @exception nsdb.ReadOnlyException Thrown if this Feature cannot be changed.
   * @exception OutOfDate if there is a an existing qualifier with the same
   *   name as the argument qualifier and it has changed since the time
   *   given by datestamp.
   **/
  public void setQualifier (nsdb.Datestamp datestamp,
                            nsdb.NucFeaturePackage.Qualifier qualifier)
      throws type.InvalidRelation, nsdb.InvalidQualifier,
             nsdb.QualifierParse, nsdb.OutOfDate, nsdb.ReadOnlyException {
    checkStamp (datestamp);
    try {
      feature.setQualifier (getQualifierFromStruct (qualifier));
      touch ();
    } catch (ReadOnlyException e) {
      throw new nsdb.ReadOnlyException ();
    } catch (InvalidRelationException e) {
      throw new InvalidRelation (e.getMessage ());
    } catch (QualifierParseException e) {
      throw new nsdb.QualifierParse ();
    } catch (InvalidQualifierException e) {
      throw new nsdb.InvalidQualifier ();
    } catch (EntryInformationException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    }
  }

  //
  // IDL:nsdb/NucFeatureWriter/removeQualiferByName:1.0
  //
  /**
   * Remove the Qualifier with the given name.  If there is no Qualifier
   * with that name, then return immediately.
   * @param name The Qualifier name to look for.
   * @exception OutOfDate if there is a an existing qualifier with the same
   *   name as the argument name and it has changed since the time given by
   *   the datestamp.
   **/
  public void removeQualiferByName (nsdb.Datestamp datestamp, String name)
      throws type.InvalidRelation, nsdb.OutOfDate, nsdb.ReadOnlyException {

  }

  //
  // IDL:nsdb/NucFeature/getNucSeqs:1.0
  //
  /**
   * If the location of the feature references multiple sequences, get a
   * sequence of all sequences referenced
   * This method is equivalent to retrieving the feature location, and looping
   *  through all location nodes to find the referenced sequences, converting
   * the accession numbers into DbXref's.
   **/
  public DbXref [] getNucSeqs () {
    return null;
  }


  //
  // IDL:seqdb/Feature/getFeatureId:1.0
  //
  /**
   * feature identifier.
   **/
  public String getFeatureId () {
    return null;
  }

  //
  // IDL:seqdb/Feature/getFeatureVersion:1.0
  //
  /**
   * return current version of the feature. returns 0 if
   * no versioning is implemented.
   **/
  public int getFeatureVersion () {
    return 0;
  }

  //
  // IDL:nsdb/NucFeatureWriter/getDatestamp:1.0
  //
  /**
   * Return a Datestamp that will be passed to the set methods on the
   * EmblSeqWriter methods and NucFeatureWriter methods.  The object that is
   * returned represents the time when the feature was last changed (the
   * last time the key, location or qualifiers changed).
   **/
  public nsdb.Datestamp getDatestamp () {
//    System.out.println ("server getDatestamp (): " + datestamp.value);
    return datestamp;
  }

  /**
   *  Return the underlying Feature object of this NucFeatureWriter.
   **/
  public Feature getDbFeature () {
    return feature;
  }

  /**
   *  Return a NucFeatureWriter object for the given Feature - this method
   *  will return the same reference every call.
   **/
  public static nsdb.NucFeatureWriter
    getNucFeatureOfFeature (final EmblSeqWriterImpl owner,
                            final Feature feat) {

    final nsdb.NucFeatureWriter user_data =
      (nsdb.NucFeatureWriter) feat.getUserData ();

    if (user_data == null) {
      final nsdb.NucFeatureWriter new_nuc_feature =
        new NucFeatureWriterImpl (owner, feat);

      feat.setUserData (new_nuc_feature);

      return new_nuc_feature;
    } else {
      return user_data;
    }
  }

  /**
   *  Update the datestamp field of this feature so that it contains the
   *  current date/time.
   **/
  private void touch () {
    datestamp = getCurrentTime ();

//    System.out.println ("server: new datestamp: " + datestamp.value);
  }

  /**
   *  Return a Datestamp object containing the current date/time.
   **/
  private nsdb.Datestamp getCurrentTime () {
    final java.util.Calendar calendar = java.util.Calendar.getInstance ();

    final java.util.Date current_time = calendar.getTime ();

    final nsdb.Datestamp datestamp = new nsdb.Datestamp ();

    datestamp.value = (int) (current_time.getTime () / 1000);

    return datestamp;
  }

  /**
   *  Given a nsdb.NucFeaturePackage.Qualifier reference return a
   *  embl.Qualifier object.
   **/
  private Qualifier
    getQualifierFromStruct (final nsdb.NucFeaturePackage.Qualifier handle)
      throws QualifierParseException, InvalidQualifierException {
    final StringVector values = getQualifierStringValues (handle);

    if (values.size () == 0) {
      return new Qualifier (handle.name);
    } else {
      return new Qualifier (handle.name, values);
    }
  }

  /**
   *  Return  the values of the given qualifier as String objects.
   **/
  private StringVector
    getQualifierStringValues (nsdb.NucFeaturePackage.Qualifier qualifier) {

    final QualifierValue_u [] corba_values =
      qualifier.values;

    final StringVector return_array = new StringVector ();

    for (int value_index = 0 ;
         value_index < corba_values.length ;
         ++value_index) {

      final QualifierValue_u corba_value =
        corba_values[value_index];

      /*final*/ String this_value_string;

      switch (corba_value.discriminator ()) {
      case nsdb.NucFeature.string_qtc:
        return_array.add (corba_value.text ());
        break;
      case nsdb.NucFeature.integer_qtc:
        return_array.add (String.valueOf (corba_value.integer ()));
        break;
      case nsdb.NucFeature.real_qtc:
        return_array.add (String.valueOf (corba_value.real ()));
        break;
      case nsdb.NucFeature.TranslationException_qtc:
        return_array.add (String.valueOf (corba_value.translation_exception ()));
        break;
      case nsdb.NucFeature.CodonTranslation_qtc:
        return_array.add (String.valueOf (corba_value.codon_translation ()));
        break;
      case nsdb.NucFeature.Anticodon_qtc:
        return_array.add (String.valueOf (corba_value.anti_codon ()));
        break;
      case nsdb.NucFeature.SpliceConsensus_qtc:
        return_array.add (String.valueOf (corba_value.splice_consensus ()));
        break;
      case nsdb.NucFeature.RepeatUnit_qtc:
        return_array.add (String.valueOf (corba_value.repeat_unit ()));
        break;
      case nsdb.NucFeature.DbXref_qtc:
        return_array.add (String.valueOf (corba_value.db_xref ()));
        break;
      }
    }

    return return_array;
  }

  /**
   *  If the Feature corresponding to this NucFeature has been deleted from
   *  it's entry, recreate it and put it back in the entry.  This is called by
   *  the set methods.
   **/
  private void maybeReAddFeature ()
      throws ReadOnlyException {
    if (!owner.entry.contains (feature)) {
      try {
        owner.entry.add (feature);
      } catch (EntryInformationException e) {
        throw new Error ("internal error - unexpected exception: " + e);
      } 
    }
  }

  /**
   *  Compare the given datestamp with the datestamp of this object and throw
   *  an OutOfDate exception if the given datestamp is older.
   **/
  private void checkStamp (final nsdb.Datestamp datestamp)
      throws nsdb.OutOfDate {
    if (datestamp.value != 0 && this.datestamp.value > datestamp.value) {
//        System.out.println ("in checkStamp (): " +
//                            datestamp.value + " " +
//                            this.datestamp.value);
      throw new nsdb.OutOfDate ();
    }
  }

  /**
   *  The Feature that was passed to the constructor.
   **/
  private Feature feature;

  /**
   *  The FeatureLocationImpl object for this feature's location.
   **/
  private FeatureLocationImpl feature_location_impl = null;

  /**
   *  The date/time this feature was last changed
   **/
  private nsdb.Datestamp datestamp;

  /**
   *  The EmblSeqWriterImpl reference of the owning entry of this feature (as
   *  passed to the constructor).
   **/
  private EmblSeqWriterImpl owner = null;
}

